/*
 * Ch37x USB driver - 1.0
 *
 * Copyright (C) 2015-2016 wch tech
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 *
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/poll.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/version.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION( ver, rel, seq )	((ver << 16) | (rel << 8) | (seq))
#endif

//#define DEBUG

#define info(arg...) do{printk(KERN_INFO arg);}while(0)
#define err(arg...) do{printk(KERN_ERR arg);}while(0)

#ifdef DEBUG
#define dbg(arg...) do{printk(KERN_DEBUG arg);}while(0)
#else
#define dbg(arg...) 
#endif

/* Define these values to match your devices */
#define USB_ch37x_VENDOR_ID_1 		0x4348
#define USB_ch37x_VENDOR_ID_2 		0x1A86
#define USB_ch37x_VENDOR_ID_3 		0x1C00
#define USB_ch37x_PRODUCT_ID_1 		0x5537
#define USB_ch37x_PRODUCT_ID_2 		0x5538
#define USB_ch37x_PRODUCT_ID_3 		0x5576

/* table of devices that work with this driver */
static struct usb_device_id ch37x_table [] = {
	{ USB_DEVICE(USB_ch37x_VENDOR_ID_1, USB_ch37x_PRODUCT_ID_1) },	
	{ USB_DEVICE(USB_ch37x_VENDOR_ID_1, USB_ch37x_PRODUCT_ID_2) },	
	{ USB_DEVICE(USB_ch37x_VENDOR_ID_3, USB_ch37x_PRODUCT_ID_1) },
	{USB_DEVICE(USB_ch37x_VENDOR_ID_2, USB_ch37x_PRODUCT_ID_3)},
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, ch37x_table);

/* Get a minor range for your devices from the usb maintainer */
#define USB_ch37x_MINOR_BASE	200

#define COMMAND_TIMEOUT	(2*HZ)	/* seconds timeout for a command */

#define WRITES_IN_FLIGHT 8 
#define READS_IN_FLIGHT  8

#define MAX_NUM 4

/* Structure to hold all of our device specific stuff */
struct usb_ch37x 
{
	struct usb_device       *udev;		/* the usb device for this device */
	struct usb_interface    *interface;	/* the interface for this device */
	struct semaphore limit_sem;		/* limiting the number of writes in progress */
	struct usb_anchor submitted;          /* in case we need to retract our submissions */
	__u16 			*ch37x_id;
	
	struct urb		*bulk_in_urb[MAX_NUM*READS_IN_FLIGHT];			/* the urb to read data with */
	unsigned char 	*bulk_in_buffer[MAX_NUM*READS_IN_FLIGHT];		/* the buffer to receive data */
	size_t			bulk_in_size[MAX_NUM];							/* the size of the receive buffer */
	__u8			bulk_in_endpointAddr[MAX_NUM];					/* the address of the bulk in endpoint */
	bool 			ongoing_read[MAX_NUM*READS_IN_FLIGHT];			/* a read is going on */
	size_t 			bulk_in_num;							/* the number of bulk in endpoint */
	
	size_t 			bulk_out_size[MAX_NUM];
	__u8			bulk_out_endpointAddr[MAX_NUM];	/* the address of the bulk out endpoint */
	size_t 			bulk_out_num;				/* the number of bulk in endpoint */

	int 				errors; 						/* the last request tanked */
	struct kref		kref;
	struct mutex 		io_mutex;

	struct RX_BUFFER_INFO {
		unsigned long Read;
		unsigned long Write;
		unsigned long Length;
		unsigned char *RxBuf;
		unsigned long Space;
		bool OverFlow;
	}RxBufInfo[MAX_NUM];

	unsigned char *read_buffer_primary[MAX_NUM];
	int			read_buffer_length[MAX_NUM];
	unsigned char *read_buffer_secondary[MAX_NUM];
	int			secondary_head[MAX_NUM];
	int			secondary_tail[MAX_NUM];
	spinlock_t	buflock;

	wait_queue_head_t read_wait[MAX_NUM];
	wait_queue_head_t write_wait[MAX_NUM];


	unsigned char		*interrupt_in_buffer[MAX_NUM];
	struct usb_endpoint_descriptor *interrupt_in_endpoint[MAX_NUM];
	struct urb	*interrupt_in_urb[MAX_NUM];
	int			read_urb_finished[MAX_NUM];
	size_t		interrupt_in_num;
	unsigned char 		interrupt_in_serial_number;

	unsigned char		*interrupt_out_buffer[MAX_NUM];
	struct usb_endpoint_descriptor *interrupt_out_endpoint[MAX_NUM];
	struct urb	*interrupt_out_urb[MAX_NUM];
	int			out_urb_finished[MAX_NUM];
	size_t		interrupt_out_num;
	
};

#define RxReset(p, num) p->RxBufInfo[num].Read = p->RxBufInfo[num].Write = 0
#define RxLength(p, num) (p->RxBufInfo[num].Length)
#define RxWrite(p, num) (p->RxBufInfo[num].Write)
#define RxRead(p, num)  (p->RxBufInfo[num].Read)
#define RxFull(p, num) (p->RxBufInfo[num].OverFlow)
#define RxSpace(p, num) (p->RxBufInfo[num].Space)
#define RxBufAvail(p, num) (RxLength(p, num) - RxSpace(p, num))
#define RxBufWrite(p, num) (p->RxBufInfo[num].RxBuf + p->RxBufInfo[num].Write) 
#define RxBufRead(p, num) (p->RxBufInfo[num].RxBuf + p->RxBufInfo[num].Read)

#define to_ch37x_dev(d) 	container_of(d, struct usb_ch37x, kref)

#define MIN(x,y)   ( (x) < (y) ? (x) : (y) )
#define DRV_VERSION		"WCH CH37x Driver Version 1.1"

#define GetVersion			0x00000001
#define GetDeviceID			0x00000003

#define GetDeviceMsg			0x00000005

#define CH37X_BULK0_UP 		0x80000000
#define CH37X_BULK1_UP 		0x80000001
#define CH37X_BULK2_UP 		0x80000002
#define CH37X_BULK3_UP 		0x80000003

#define CH37X_BULK0_DOWN 	0x80000004
#define CH37X_BULK1_DOWN 	0x80000005
#define CH37X_BULK2_DOWN 	0x80000006
#define CH37X_BULK3_DOWN 	0x80000007

#define CH37X_INT0_UP 		0x80000008
#define CH37X_INT1_UP 		0x80000009
#define CH37X_INT2_UP 		0x8000000a
#define CH37X_INT3_UP 		0x8000000b

#define CH37X_INT0_DOWN 	0x8000000c
#define CH37X_INT1_DOWN 	0x8000000d
#define CH37X_INT2_DOWN 	0x8000000e
#define CH37X_INT3_DOWN 	0x8000000f

static struct usb_driver ch37x_driver;

static void ch37x_draw_down(struct usb_ch37x *dev);

static void ch37x_interrupt_in_callback(struct urb *urb)
{
	struct usb_ch37x *dev = urb->context;
	int status = urb->status;
	int num = 0;

	for(num=0; num<MAX_NUM; num++) {
		if(urb->pipe == usb_rcvintpipe(dev->udev, dev->interrupt_in_endpoint[num]->bEndpointAddress))
		break;
	}

	dbg("%s is excuting, num: %d\n", __func__, num);

	spin_lock(&dev->buflock);

	if (status != 0) {
		if ((status != -ENOENT) && (status != -ECONNRESET) &&
			(status != -ESHUTDOWN)) {
		}
		err("%s: nonzero status received: %d", __func__, status);
		goto exit;
	}

	dbg("urb->actual_length: %d\n", urb->actual_length);
	if (urb->actual_length) {
		if(dev->read_buffer_length[num] <
			(4 * le16_to_cpu(dev->interrupt_in_endpoint[num]->wMaxPacketSize)) -
			(urb->actual_length)) {		
			memcpy(dev->read_buffer_primary[num] +
				dev->read_buffer_length[num],
				dev->interrupt_in_buffer[num], urb->actual_length);

			dev->read_buffer_length[num] += urb->actual_length;
			dbg("%s reading  %d\n", __func__, urb->actual_length);
		} else {
			dbg("%s: read_buffer overflow\n", __func__);
		}
	} else {
		dbg("%s: read_buffer null\n", __func__);
	}

exit:
	dev->read_urb_finished[num] = 1;
	spin_unlock(&dev->buflock);
	/* always wake up so we recover from errors */
	wake_up_interruptible(&dev->read_wait[num]);
}

static void ch37x_interrupt_out_callback(struct urb *urb)
{
	struct usb_ch37x *dev = urb->context;
	int status = urb->status;
	int num = 0;

	if (status != 0) {
		if ((status != -ENOENT) &&
		    (status != -ECONNRESET)) {
		}
		err("%s :nonzero status received: %d\n", __func__, status);
		return;
	}

	for(num = 0; num<MAX_NUM; num++) {
		if(urb->pipe == usb_sndintpipe(dev->udev, dev->interrupt_out_endpoint[num]->bEndpointAddress))
		break;
	}

	spin_lock(&dev->buflock);
	dev->out_urb_finished[num] = 1;
	wake_up(&dev->write_wait[num]);
	spin_unlock(&dev->buflock);
}

static void ch37x_bulk_in_callback(struct urb *urb)
{
	struct usb_ch37x *dev;
	int num = 0;
	int ep_num = 0;
	int i, j;

	dev = urb->context;

	for(num = 0; num < MAX_NUM*READS_IN_FLIGHT; num++) {
		if(urb == dev->bulk_in_urb[num])
			break;
	}
	ep_num = num / READS_IN_FLIGHT;

	dbg("%s-->ep_num: %d, num: %d\n", __func__, ep_num, num);

	spin_lock(&dev->buflock);

	/* sync/async unlink faults aren't errors */
	dbg("urb[%d]->status: %d\n", num, urb->status);
	if(urb->status) {
		if (!(urb->status == -ENOENT ||
		    urb->status == -ECONNRESET ||
		    urb->status == -ESHUTDOWN))
			err("%s - nonzero write bulk status received: %d\n",
				__func__, urb->status);
		dev->errors = urb->status;
	} else {
		static int first, second;
		
		for(i = 0; i < urb->actual_length; i++) {
			if(dev->bulk_in_buffer[num][i] != i % 256) {
				for(j = 0; j < urb->actual_length; j++)
					info("0x%.2x ", dev->bulk_in_buffer[num][j]);
			}
		}

		dbg("1.Space: %ld\n", RxSpace(dev, ep_num));
		dbg("1.Read: %ld\n", RxRead(dev, ep_num));
		dbg("1.Write: %ld\n", RxWrite(dev, ep_num));
		dbg("1.RxFull: %d\n", RxFull(dev, ep_num));

		if(RxWrite(dev, ep_num) >= RxRead(dev, ep_num)) {
			first = RxLength(dev, ep_num) - RxWrite(dev, ep_num);
			dbg("first: %d\n", first);
			dbg("urb->actual_length: %d\n", urb->actual_length);
			if(first >= urb->actual_length) {
				memcpy(RxBufWrite(dev, ep_num), dev->bulk_in_buffer[num], 
						urb->actual_length);
				RxWrite(dev, ep_num) += urb->actual_length;
				RxSpace(dev, ep_num) = RxLength(dev, ep_num) - 
					(RxWrite(dev, ep_num) - RxRead(dev, ep_num));
			} else {
				dbg("second!\n");
				second = urb->actual_length - first;
				memcpy(RxBufWrite(dev, ep_num), dev->bulk_in_buffer[num],
						first);
				RxWrite(dev, ep_num) = 0;
				memcpy(RxBufWrite(dev, ep_num), dev->bulk_in_buffer[num] +
						first, second);
				RxWrite(dev, ep_num) = second;
				RxSpace(dev, ep_num) = RxRead(dev, ep_num) - RxWrite(dev, ep_num);
			}
			if(RxSpace(dev, ep_num) < (READS_IN_FLIGHT * dev->bulk_in_size[ep_num])) {
				RxFull(dev, ep_num) = 1;
			} else
				RxFull(dev, ep_num) = 0;
		} else {
			dbg("third!\n");
			memcpy(RxBufWrite(dev, ep_num), dev->bulk_in_buffer[num], 
					urb->actual_length);
			RxWrite(dev, ep_num) += urb->actual_length;
			RxSpace(dev, ep_num) = RxRead(dev, ep_num) - RxWrite(dev, ep_num);
			if(RxSpace(dev, ep_num) < (READS_IN_FLIGHT * dev->bulk_in_size[ep_num])) {
				RxFull(dev, ep_num) = 1;
			} else
				RxFull(dev, ep_num) = 0;
			dbg("4.Space: %ld\n", RxSpace(dev, ep_num));
			dbg("4.Read: %ld\n", RxRead(dev, ep_num));
			dbg("4.Write: %ld\n", RxWrite(dev, ep_num));
			dbg("4.RxFull: %d\n", RxFull(dev, ep_num));
		}
		RxWrite(dev, ep_num) %= RxLength(dev, ep_num);

		dbg("2.Space: %ld\n", RxSpace(dev, ep_num));
		dbg("2.Read: %ld\n", RxRead(dev, ep_num));
		dbg("2.Write: %ld\n", RxWrite(dev, ep_num));
		dbg("2.RxFull: %d\n", RxFull(dev, ep_num));
	}

	dev->ongoing_read[num] = 0;
	spin_unlock(&dev->buflock);
}

static void ch37x_bulk_out_callback(struct urb *urb)
{
	struct usb_ch37x *dev;

	dbg("%s excuting!\n", __func__);
	
	dev = urb->context;

	/* sync/async unlink faults aren't errors */
	if(urb->status) {
	if (!(urb->status == -ENOENT || 
	      urb->status == -ECONNRESET ||
	      urb->status == -ESHUTDOWN)) 
			dbg("%s - nonzero write bulk status received: %d",
		    	__FUNCTION__, urb->status);

		spin_lock(&dev->buflock);
		dev->errors = urb->status;
		spin_unlock(&dev->buflock);
	}
	/* free up our allocated buffer */
	usb_free_coherent(urb->dev, urb->transfer_buffer_length,
			urb->transfer_buffer, urb->transfer_dma);
	up(&dev->limit_sem);
}

static ssize_t ch37x_int_up(struct usb_ch37x *dev, int num, unsigned long pbuf, size_t count)
{
	size_t bytes_read = 0;
	size_t bytes_to_read = count;
	int i;
	int retval = 0;
	int timeout = 0;
	int should_submit = 0;
	unsigned long flags;
	DECLARE_WAITQUEUE(wait, current);

	/* verify that some data was requested */
	if (count == 0) {
		dbg("%s : read request of 0 bytes\n", __func__);
		goto exit;
	}

	timeout = COMMAND_TIMEOUT;
	dbg("%s : about to start looping", __func__);
	while (bytes_to_read) {
		int data_in_secondary = dev->secondary_tail[num] - dev->secondary_head[num];
		dbg("%s : while, data_in_secondary=%d, status=%d\n",
			__func__, data_in_secondary,
			dev->interrupt_in_urb[num]->status);

		if (data_in_secondary) {
			/* drain secondary buffer */
			int amount = bytes_to_read < data_in_secondary ? bytes_to_read : data_in_secondary;
			i = copy_to_user((char __user *)pbuf, dev->read_buffer_secondary[num]+dev->secondary_head[num], amount);
			if(i) {
				retval = -EFAULT;
				goto exit;
			}
			dev->secondary_head[num] += amount;
			pbuf += amount;
			bytes_read += amount;
			bytes_to_read -= amount;
		} else {
			/* we check the primary buffer */
			spin_lock_irqsave (&dev->buflock, flags);
			if (dev->read_buffer_length[num]) {
				/* we secure access to the primary */
				unsigned char *tmp;
				dbg("%s : swap, read_buffer_length = %d\n", __func__, dev->read_buffer_length[num]);
				tmp = dev->read_buffer_secondary[num];
				dev->read_buffer_secondary[num] = dev->read_buffer_primary[num];
				dev->read_buffer_primary[num] = tmp;
				dev->secondary_head[num] = 0;
				dev->secondary_tail[num] = dev->read_buffer_length[num];
				dev->read_buffer_length[num] = 0;
				spin_unlock_irqrestore(&dev->buflock, flags);
				/* we have a free buffer so use it */
				should_submit = 1;
			} else {
				/* even the primary was empty - we may need to do IO */
				if (!dev->read_urb_finished[num]) {
					/* somebody is doing IO */
					spin_unlock_irqrestore(&dev->buflock, flags);
					dbg("%s : submitted already", __func__);
				} else {
					/* we must initiate input */
					dbg("%s : initiate input\n", __func__);
					dev->read_urb_finished[num] = 0;
					spin_unlock_irqrestore(&dev->buflock, flags);
					retval = usb_submit_urb(dev->interrupt_in_urb[num], GFP_KERNEL);
					if (retval) {
						dev->read_urb_finished[num] = 1;
						if (retval == -ENOMEM) {
							retval = bytes_read ? bytes_read : -ENOMEM;
						}
						dbg("%s : submit failed\n", __func__);
						goto exit;
					}
				}

				/* we wait for I/O to complete */
				set_current_state(TASK_INTERRUPTIBLE);
				add_wait_queue(&dev->read_wait[num], &wait);
				spin_lock_irqsave(&dev->buflock, flags);
				if (!dev->read_urb_finished[num]) {
					spin_unlock_irqrestore(&dev->buflock, flags);
					dbg("schedule_timeout!\n");
					timeout = schedule_timeout(COMMAND_TIMEOUT);
				} else {
					spin_unlock_irqrestore(&dev->buflock, flags);
					set_current_state(TASK_RUNNING);
				}
				remove_wait_queue(&dev->read_wait[num], &wait);

				if (timeout <= 0) {
					dbg("%s : timeout\n", __func__);
					retval = bytes_read ? bytes_read : -ETIMEDOUT;
					goto exit;
				}
				dbg("timeout: %d\n", timeout);

				if (signal_pending(current)) {
					dbg("%s : signal pending\n", __func__);
					retval = bytes_read ? bytes_read : -EINTR;
					goto exit;
				}
			} /* end of read_buffer_length... */
		} /* end of data_in_secondary */
	} /* end of while */

	retval = bytes_read;
	/* if the primary buffer is empty then use it */
	spin_lock_irqsave(&dev->buflock, flags);
	if (should_submit && dev->read_urb_finished[num]) {
		dev->read_urb_finished[num] = 0;
		spin_unlock_irqrestore(&dev->buflock, flags);
		if (usb_submit_urb(dev->interrupt_in_urb[num], GFP_KERNEL) != 0)
			dev->read_urb_finished[num] = 1;
		dbg("%s resubmit urb!\n", __func__);
		/* we ignore failure */
	} else {
		spin_unlock_irqrestore(&dev->buflock, flags);
	}

exit:
	return retval;
}

static ssize_t ch37x_int_down(struct usb_ch37x *dev, int num, unsigned long pbuf, size_t count)
{
	size_t bytes_written = 0;
	size_t bytes_to_write;
	size_t buffer_size;
	unsigned long flags;
	int retval = 0;

	/* init wait queue head */
	DECLARE_WAITQUEUE(waita, current);		

	/* verify that we actually have some data to write */
	if (count == 0) {
		dbg("%s : write request of 0 bytes\n", __func__);
		goto exit;
	}

	while (count > 0) {
		/* add wait queue 'waita' to wait queue head */
		add_wait_queue(&dev->write_wait[num], &waita);
		/* be waken up by signal */
		set_current_state(TASK_INTERRUPTIBLE);			
		spin_lock_irqsave(&dev->buflock, flags);		
		if (!dev->out_urb_finished[num]) {
			spin_unlock_irqrestore(&dev->buflock, flags);	
			mutex_unlock(&dev->io_mutex);			
			/* check if there signal to deal with(nonzero if signal comes), if
			 * signal valid, no need to sleep now 
			 */
			if (signal_pending(current)) {			
				dbg("%s : interrupted\n", __func__);
				/* move process to running queue and run when being scheduling */
				set_current_state(TASK_RUNNING);	 
				/* Interrupted system call */
				retval = -EINTR;				
				goto exit_onqueue;
			}
			/* 
			 * sleep until timeout or signal 
			 * @return: 0 ==> be waken up cause timeout
			 * 	 		else ==> remain time
			 */	
			 if (schedule_timeout(COMMAND_TIMEOUT) == 0) {	
				dbg("%s - command timed out.\n", __func__);
				retval = -ETIMEDOUT;
				goto exit_onqueue;
			}
			/* if process be waken up by wake_up */
			/* remove wait queue 'waita' from wait queue head */
			remove_wait_queue(&dev->write_wait[num], &waita);	  
			/* 
			 * return 0 if get mutex, if a signal arrives while waiting
			 * for the lock then this function returns -EINTR.
			 */
			retval = mutex_lock_interruptible(&dev->io_mutex);    
			if (retval) {
				retval = bytes_written ? bytes_written : retval;
				goto exit_nolock;
			}
			dbg("%s : in progress, count = %Zd\n", __func__, count);
		} else {
			spin_unlock_irqrestore(&dev->buflock, flags);			
			set_current_state(TASK_RUNNING);
			remove_wait_queue(&dev->write_wait[num], &waita);
			dbg("%s : sending, count = %Zd\n", __func__, count);

			/* write the data into interrupt_out_buffer from userspace */
			buffer_size = le16_to_cpu(dev->interrupt_out_endpoint[num]->wMaxPacketSize);
			bytes_to_write = count > buffer_size ? buffer_size : count;
			dbg("%s : buffer_size = %Zd, count = %Zd, bytes_to_write = %Zd\n",
				__func__, buffer_size, count, bytes_to_write);

			if (copy_from_user(dev->interrupt_out_buffer[num], (char __user *)pbuf, bytes_to_write) != 0) {
				retval = -EFAULT;
				goto exit;
			}

			/* send off the urb */
			usb_fill_int_urb(
				dev->interrupt_out_urb[num],
				dev->udev,
				usb_sndintpipe(dev->udev, dev->interrupt_out_endpoint[num]->bEndpointAddress),
				dev->interrupt_out_buffer[num],
				bytes_to_write,
				ch37x_interrupt_out_callback,
				dev,
				dev->interrupt_out_endpoint[num]->bInterval);
			dev->interrupt_out_urb[num]->actual_length = bytes_to_write;
			dev->out_urb_finished[num] = 0;
			retval = usb_submit_urb(dev->interrupt_out_urb[num], GFP_KERNEL);
			if (retval < 0) {
				dev->out_urb_finished[num] = 1;
				dbg("Couldn't submit interrupt_out_urb %d\n", retval);
				goto exit;
			}
			pbuf += bytes_to_write;
			count -= bytes_to_write;
			bytes_written += bytes_to_write;
		}
	}
	return bytes_written;

exit:
exit_nolock:
	return retval;
exit_onqueue:
	remove_wait_queue(&dev->write_wait[num], &waita);
	return retval;
}

static int ch37x_do_read_io(struct usb_ch37x *dev, int ep_num, size_t count)
{
	int rv = 0;
	int num = 0;
	bool ongoing_io;
	int num_start = READS_IN_FLIGHT* ep_num;
	int num_end =  READS_IN_FLIGHT * (ep_num + 1);
	
	spin_lock_irq(&dev->buflock);
	if(RxSpace(dev, ep_num) >= READS_IN_FLIGHT * dev->bulk_in_size[ep_num]) {
		RxFull(dev, ep_num) = 0;
	} else
		RxFull(dev, ep_num) = 1;
	spin_unlock_irq(&dev->buflock);

	if(RxFull(dev, ep_num))
		return 0;

	for(num = num_start; num < num_end; num++) {

		spin_lock_irq(&dev->buflock);
		ongoing_io = dev->ongoing_read[num];
		spin_unlock_irq(&dev->buflock);

		if(ongoing_io)
			continue;

		/* prepare a read */
		usb_fill_bulk_urb(dev->bulk_in_urb[num],
				dev->udev,
				usb_rcvbulkpipe(dev->udev,
					dev->bulk_in_endpointAddr[ep_num]),
				dev->bulk_in_buffer[num],
				dev->bulk_in_size[ep_num], 
				ch37x_bulk_in_callback,
				dev);

		spin_lock_irq(&dev->buflock);
		dev->ongoing_read[num] = 1;
		spin_unlock_irq(&dev->buflock);
		/* do it */
		rv = usb_submit_urb(dev->bulk_in_urb[num], GFP_ATOMIC);
		if (rv < 0) {
			err("%s - failed submitting read urb, error %d\n",
					__func__, rv);
			rv = (rv == -ENOMEM) ? rv : -EIO;
			spin_lock_irq(&dev->buflock);
			dev->ongoing_read[num] = 0;
			spin_unlock_irq(&dev->buflock);
		} else
			dbg("%s - urb[%d] submit success!\n", __func__, num);
	}
	return rv;
}

static ssize_t ch37x_bulk_up(struct usb_ch37x *dev, int num, unsigned long pbuf, unsigned long count)
{
	int rv;
	ssize_t bytes_read = 0;
	ssize_t bytes_to_read = count;
	int amount;
	unsigned char *buffer = NULL;

	buffer = kmalloc(READS_IN_FLIGHT * dev->bulk_in_size[num], GFP_KERNEL);
	if(!buffer) {
		err("buffer fail!\n");
		rv = -ENOMEM;
		goto exit;	
	}

	/* if we cannot read at all, return EOF */
	if (!dev->bulk_in_urb[0] || !count)
		return 0;

	/* no concurrent readers */
	if (!dev->interface) {		/* disconnect() was called */
		rv = -ENODEV;
		goto exit;
	}

	/* errors must be reported */
	rv = dev->errors;
	if (rv < 0) {
		/* any error is reported once */
		dev->errors = 0;
		/* to preserve notifications about reset */
		rv = (rv == -EPIPE) ? rv : -EIO;
		err("error happend!\n");
		/* report it */
		goto exit;
	}

	/*
	 * if the buffer is filled we may satisfy the read
	 * else we need to start IO
	 */
	while(bytes_to_read) {
		spin_lock_irq(&dev->buflock);
		if(RxBufAvail(dev, num)) {

			amount = MIN(RxBufAvail(dev, num), RxLength(dev, num) - RxRead(dev, num));
			amount = MIN(amount, bytes_to_read);
			memcpy(buffer, RxBufRead(dev, num), amount);

			memset(RxBufRead(dev, num), 0, amount);

			RxRead(dev, num) += amount;
			RxRead(dev, num) %= RxLength(dev, num);

			RxSpace(dev, num) += amount;

			dbg("----------------------------------------------------->\n");
			dbg("%s - amount: %d\n", __func__, amount);
			dbg("%s - bytes_read: %d\n", __func__, bytes_read);
			dbg("3.Space: %ld\n", RxSpace(dev, num));
			dbg("3.Read: %ld\n", RxRead(dev, num));
			dbg("3.Write: %ld\n", RxWrite(dev, num));
			dbg("3.RxFull: %d\n", RxFull(dev, num));

			spin_unlock_irq(&dev->buflock);

			if(copy_to_user((char __user *)pbuf, buffer, amount)) {
				err("copy_to_user error!\n");
				/* if copy fail, we should recovery the changes */
				RxSpace(dev, num) -= amount;
				if(RxRead(dev, num) < amount)
					RxRead(dev, num) = RxRead(dev, num) + RxLength(dev, num) - amount;
				else
					RxRead(dev, num) -= amount;
				memcpy(RxBufRead(dev, num), buffer, amount);	 
				rv = -EFAULT;
				goto exit;
			}
			pbuf += amount;
			bytes_read += amount;
			bytes_to_read -= amount;

		} else {
			spin_unlock_irq(&dev->buflock);
			break;
		}
	}
	kfree(buffer);	
	rv = ch37x_do_read_io(dev, num, count);
	if (rv < 0)
		goto exit;

	return bytes_read;
exit:
	return rv;
}

static ssize_t ch37x_bulk_down(struct usb_ch37x *dev, int num, unsigned long pbuf, unsigned long count)
{
	int retval = 0;
	struct urb *urb = NULL;
	char *buf = NULL;
	size_t bytes_written = 0;
	size_t bytes_to_write = 0;
	size_t buffer_size = dev->bulk_out_size[num];

	/* verify that we actually have some data to write */
	if (count == 0)
		goto exit;

	while(count > 0) {

		if(down_interruptible(&dev->limit_sem)) {
			retval = -ERESTARTSYS;
			goto exit;
		}

		spin_lock_irq(&dev->buflock);
		retval = dev->errors;
		if(retval < 0) {
			dev->errors = 0;
			retval = (retval == -EPIPE) ? retval : -EIO;
		}
		spin_unlock_irq(&dev->buflock);
		if(retval < 0)
			goto error;

		bytes_to_write = count > buffer_size ? buffer_size : count;

		urb = usb_alloc_urb(0, GFP_KERNEL);
		if(!urb) {
			retval = -ENOMEM;
			goto error;
		}

		buf = usb_alloc_coherent(dev->udev, bytes_to_write, 
				GFP_KERNEL, &urb->transfer_dma);
		if(!buf) {
			retval = -ENOMEM;
			goto error;
		}

		if (copy_from_user(buf, (long __user *)pbuf, bytes_to_write)) {
			retval = -EFAULT;
			goto error;
		}

		/* initialize the urb properly */
		usb_fill_bulk_urb(urb, dev->udev,
				usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr[num]),
				buf, bytes_to_write, ch37x_bulk_out_callback, dev);
		urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
		usb_anchor_urb(urb, &dev->submitted);

		/* send the data out the bulk port */
		retval = usb_submit_urb(urb, GFP_KERNEL);
		if (retval) {
			err("%s - failed submitting write urb, error %d\n", __FUNCTION__, retval);
			goto error_unanchor;
		}
		else
			dbg("submit success!\n");

		pbuf += bytes_to_write;
		count -= bytes_to_write;
		bytes_written += bytes_to_write;

		/* release our reference to this urb, the USB core will eventually free it entirely */
		usb_free_urb(urb);
	}

	return bytes_written;

error_unanchor:
	dbg("error_unanchor!\n");
	usb_unanchor_urb(urb);

error:
	if(urb) {
		usb_free_coherent(dev->udev, bytes_to_write, buf, urb->transfer_dma);
		usb_free_urb(urb);
	}
	up(&dev->limit_sem);

exit:
	return retval;

}

static void ch37x_delete(struct kref *kref)
{	
	struct usb_ch37x *dev = to_ch37x_dev(kref);
	int i = 0;

	info("%s excuting\n", __func__);

	usb_put_dev(dev->udev);

	kfree(dev->ch37x_id);

	/* free data structures */
	for( i = 0; i < READS_IN_FLIGHT * dev->bulk_in_num; i++ ) { 
		usb_free_urb(dev->bulk_in_urb[i]);
		kfree(dev->bulk_in_buffer[i]);
	}

	for(i=0; i<dev->interrupt_in_num; i++) {
		usb_free_urb(dev->interrupt_in_urb[i]);
		kfree(dev->interrupt_in_buffer[i]);
		kfree(dev->read_buffer_primary[i]);
		kfree(dev->read_buffer_secondary[i]);
	}

	for(i=0; i<dev->interrupt_out_num; i++) {
		usb_free_urb(dev->interrupt_out_urb[i]);
		kfree(dev->interrupt_out_buffer[i]);
	}
	kfree(dev);
}

static int ch37x_open(struct inode *inode, struct file *file)
{
	struct usb_ch37x *dev;
	struct usb_interface *interface;
	int subminor;
	int retval = 0;
	int i;

	info("%s excuting!\n", __func__);
	subminor = iminor(inode);

	interface = usb_find_interface(&ch37x_driver, subminor);
	if (!interface)  {
		err ("%s - error, can't find device for minor %d\n",
		     __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata(interface);
	if (!dev) 
	{
		retval = -ENODEV;
		goto exit;
	}

	retval = usb_autopm_get_interface(interface);
	if (retval)
		goto exit;

	/* increment our usage count for the device */
	kref_get(&dev->kref);

	/* save our object in the file's private structure */
	file->private_data = dev;

	if(dev->interrupt_in_num) {
		for(i=0; i<dev->interrupt_in_num; i++) {
			// initialize in direction 
			dev->read_buffer_length[i] = 0;
			dev->read_urb_finished[i] = 0;
			
			// we ignore failure 
			// end of fixup for first read 
		}
	}

	if(dev->interrupt_out_num) {
		for(i=0; i<dev->interrupt_out_num; i++) {
			// initialize out direction 
			dev->out_urb_finished[i] = 1;
		}
	}

exit:
	return retval;
}
	

static void ch37x_release_internal(struct usb_ch37x *dev)
{
	unsigned long flags;
	int i = 0;

	if(dev->udev == NULL)
		return;

	info("%s - excuting!\n", __func__);

	for(i=0; i<dev->bulk_in_num; i++) {
		spin_lock_irqsave(&dev->buflock, flags);
		if(dev->ongoing_read[i]) {
			spin_unlock_irqrestore(&dev->buflock, flags);
			dev->ongoing_read[i] = 0;
			usb_kill_urb(dev->bulk_in_urb[i]);
		} else
			spin_unlock_irqrestore(&dev->buflock, flags);
	}
	
	for(i=0; i<dev->interrupt_in_num; i++) {
		spin_lock_irqsave(&dev->buflock, flags);
		if(!dev->read_urb_finished[i]) {
			spin_unlock_irqrestore(&dev->buflock, flags);
			usb_kill_urb(dev->interrupt_in_urb[i]);
		} else
			spin_unlock_irqrestore(&dev->buflock, flags);
	}
	
	for(i=0; i<dev->interrupt_out_num; i++) {
		spin_lock_irqsave(&dev->buflock, flags);
		if(!dev->out_urb_finished[i]) {
			spin_unlock_irqrestore(&dev->buflock, flags);
			usb_kill_urb(dev->interrupt_out_urb[i]);	
		} else
			spin_unlock_irqrestore(&dev->buflock, flags);
	}
}

static int ch37x_release(struct inode *inode, struct file *file)
{
	struct usb_ch37x *dev;
	int retval = 0;

	if(file == NULL) {
		return -ENODEV;
		goto exit;
	}

	dev = (struct usb_ch37x *)file->private_data;
	if(dev == NULL) {
		retval = -ENODEV;
		goto exit;
	}

	info("%s - excuting!\n", __func__);
	
	mutex_lock(&dev->io_mutex);
	if (dev->interface)
		usb_autopm_put_interface(dev->interface);
	ch37x_release_internal(dev);
	
	mutex_unlock(&dev->io_mutex);

	/* decrement the count on our device */
	kref_put(&dev->kref, ch37x_delete);

	return 0; 

exit: 
	return retval;
}

static int ch37x_flush(struct file *file, fl_owner_t id)
{
	struct usb_ch37x *dev;
	int res;

	info("%s - excuting!\n", __func__);
	dev = file->private_data;
	if (dev == NULL)
		return -ENODEV;

	dbg("4.Space: %ld\n", RxSpace(dev, 0));
	dbg("4.Read: %ld\n", RxRead(dev, 0));
	dbg("4.Write: %ld\n", RxWrite(dev, 0));
	dbg("4.RxFull: %d\n", RxFull(dev, 0));

	/* wait for io to stop */
	mutex_lock(&dev->io_mutex);
	ch37x_draw_down(dev);

	/* read out errors, leave subsequent opens a clean slate */
	spin_lock_irq(&dev->buflock);
	res = dev->errors ? (dev->errors == -EPIPE ? -EPIPE : -EIO) : 0;
	dev->errors = 0;
	spin_unlock_irq(&dev->buflock);

	mutex_unlock(&dev->io_mutex);

	dbg("4.Space: %ld\n", RxSpace(dev, 0));
	dbg("4.Read: %ld\n", RxRead(dev, 0));
	dbg("4.Write: %ld\n", RxWrite(dev, 0));
	dbg("4.RxFull: %d\n", RxFull(dev, 0));

	return res;
}


/********************************************************************************************/

static ssize_t ch37x_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
{
	return 0;
}


static ssize_t ch37x_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
	struct usb_ch37x *dev;
	int retval = 0;
	struct urb *urb = NULL;
	char *buf = NULL;


	dbg("%s excuting!\n", __func__);


	dev = (struct usb_ch37x *)file->private_data;

	/* verify that we actually have some data to write */
	if (count == 0)
		goto exit;

	/* create a urb, and a buffer for it, and copy the data to the urb */
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!urb) {
		retval = -ENOMEM;
		goto error;
	}

#if( LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35))
	buf = usb_buffer_alloc(dev->udev, count, GFP_KERNEL, &urb->transfer_dma);
#else
	buf = usb_alloc_coherent(dev->udev, count, GFP_KERNEL, &urb->transfer_dma);
#endif
	if (!buf) {
		retval = -ENOMEM;
		goto error;
	}

	if (copy_from_user(buf, user_buffer, count)) {
		retval = -EFAULT;
		goto error;
	}

	/* initialize the urb properly */
	usb_fill_bulk_urb(urb, dev->udev,
			  usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr[0]),
			  buf, count, ch37x_bulk_out_callback, dev);
	urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	/* send the data out the bulk port */
	retval = usb_submit_urb(urb, GFP_KERNEL);
	if (retval) {
		err("%s - failed submitting write urb, error %d\n", __FUNCTION__, retval);
		goto error;
	} else 
		dbg("urb submit success!\n");

	/* release our reference to this urb, the USB core will eventually free it entirely */
	usb_free_urb(urb);

exit:
	return count;

error:
#if( LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 35))
	usb_buffer_free(dev->udev, count, buf, urb->transfer_dma);
#else
	usb_free_coherent(dev->udev, count, buf, urb->transfer_dma);
#endif
	usb_free_urb(urb);
#ifdef DEBUG
	err("%s retval: %d\n", __func__, retval);
#endif
	return retval;
}

long ch37x_ioctl(struct file *ch37x_file, unsigned int cmd, unsigned long arg)
{
	int retval = 0;
	int bytes_write = 0;
	int bytes_read = 0;
	int ep_num = 0;
	struct usb_ch37x *dev;
	char *ch37x_version_tmp = DRV_VERSION;
	
	dev = (struct usb_ch37x *)ch37x_file->private_data;

	if (mutex_lock_interruptible(&dev->io_mutex))
		return -ERESTARTSYS;

	/* verify that the device wasn't unplugged */
	if (dev->udev == NULL) {
		retval = -ENODEV;
		dbg("No device or device unplugged %d\n", retval);
		goto exit;
	}
	
	switch(cmd) {
	case GetVersion:
		retval = copy_to_user((char __user *)arg, ch37x_version_tmp,
			strlen(DRV_VERSION));
		dbg("Get Driver Version-->\n");
		break;
	case GetDeviceID:
		retval = __copy_to_user((__u16 __user *)arg, dev->ch37x_id, 
			sizeof(__u16) * 2);
		dbg("Get Device ID-->\n");
		break;
	case GetDeviceMsg:
		ep_num = (int)((((char)(dev->bulk_in_num))<<24) | (((char)(dev->bulk_out_num))<<16)  | 
				(((char)(dev->interrupt_in_num))<<8)  | (((char)(dev->interrupt_out_num))<<24) );
		retval = copy_to_user((long __user *)arg, &ep_num, sizeof(ep_num));
		dbg("Get Device Msg-->\n");
		break;
	case CH37X_BULK0_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK0_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_bulk_up(dev, 0, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_BULK0_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK0_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_bulk_down(dev, 0, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;
	case CH37X_INT0_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("INT0_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_int_up(dev, 0, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_INT0_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("INT0_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_int_down(dev, 0, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;

	case CH37X_BULK1_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK1_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_bulk_up(dev, 1, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_BULK1_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK1_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_bulk_down(dev, 1, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;
	case CH37X_INT1_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("INT1_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_int_up(dev, 1, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_INT1_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("INT1_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_int_down(dev, 1, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;

	case CH37X_BULK2_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK2_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_bulk_up(dev, 2, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_BULK2_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK2_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_bulk_down(dev, 2, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;
	case CH37X_INT2_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("INT2_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_int_up(dev, 2, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_INT2_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("INT2_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_int_down(dev, 2, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;

	case CH37X_BULK3_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK3_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_bulk_up(dev, 3, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_BULK3_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("BULK3_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_bulk_down(dev, 3, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;
	case CH37X_INT3_UP:
		retval = get_user(bytes_read, (int __user *)(*((unsigned long *)arg)));
		dbg("INT3_UP-->bytes_read: %d\n", bytes_read);
		retval = ch37x_int_up(dev, 3, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_read);
		break;
	case CH37X_INT3_DOWN:
		retval = get_user(bytes_write, (int __user *)(*((unsigned long *)arg)));
		dbg("INT3_DOWN-->bytes_write: %d\n", bytes_write);
		retval = ch37x_int_down(dev, 3, (unsigned long)((char *)(*((unsigned long *)arg + 1))), bytes_write);
		break;
		
	default:
		retval =  -ENOTTY;
		break;
	}	
exit:
	/* unlock the device */
	mutex_unlock(&dev->io_mutex);

	return retval;
}

static struct file_operations ch37x_fops = {
	.owner =	THIS_MODULE,
	.read =		ch37x_read,
	.write =	ch37x_write,
	.open =		ch37x_open,
	.unlocked_ioctl=	ch37x_ioctl,
	.release =	ch37x_release,
	.flush = ch37x_flush,
	.llseek = noop_llseek,
};

/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with devfs and the driver core
 */
static struct usb_class_driver ch37x_class = 
{
	.name =		"usb/ch37x%d",
	.fops =		&ch37x_fops,
	.minor_base =	USB_ch37x_MINOR_BASE,
};

static int ch37x_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_ch37x *dev;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	size_t buffer_size;
	int i;
	int j;
	int bulk_in_num = 0;
	int bulk_out_num = 0;
	int int_in_num = 0;
	int int_out_num = 0;
	int in_end_size;
	int out_end_size;
	int retval = -ENOMEM;
	
	/* allocate memory for our device state and initialize it */
	dev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		err("Out of memory!\n");
		goto error;
	}
	memset(dev, 0x00, sizeof(*dev));
	kref_init(&dev->kref);
	sema_init(&dev->limit_sem, WRITES_IN_FLIGHT);

	mutex_init(&dev->io_mutex);
	spin_lock_init(&dev->buflock);
	init_usb_anchor(&dev->submitted);
	dev->ch37x_id = kmalloc(sizeof(char) * 4, GFP_KERNEL);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->ch37x_id[0] = dev->udev->descriptor.idVendor;
	dev->ch37x_id[1] = dev->udev->descriptor.idProduct; 
	
	dev->interface = interface;

	/* set up the endpoint information */
	/* use only the first bulk-in and bulk-out endpoints */
	iface_desc = interface->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;
		/* USB DIR IN && BULK */
		if ((endpoint->bEndpointAddress & 0x80)  &&  ((endpoint->bmAttributes & 3) == 0x02)) {		
			/* we found a bulk in endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			dev->bulk_in_size[bulk_in_num] = buffer_size;
			dev->bulk_in_endpointAddr[bulk_in_num] = endpoint->bEndpointAddress;
			
			dev->RxBufInfo[bulk_in_num].Length = READS_IN_FLIGHT * buffer_size * 10;
			dev->RxBufInfo[bulk_in_num].RxBuf = kmalloc(RxLength(dev, bulk_in_num), GFP_KERNEL);
			if(!dev->RxBufInfo[bulk_in_num].RxBuf) {
				err("Out of memory!\n");
				goto error;
			}
			
			RxReset(dev, bulk_in_num);
			RxSpace(dev, bulk_in_num) = RxLength(dev, bulk_in_num);
			RxFull(dev, bulk_in_num) = 0;

			for(j = 0; j < READS_IN_FLIGHT; j++) {	
				dev->ongoing_read[bulk_in_num * READS_IN_FLIGHT + j] = 0;
				dev->bulk_in_urb[bulk_in_num * READS_IN_FLIGHT + j] = 
					usb_alloc_urb(0, GFP_KERNEL);
				if (!dev->bulk_in_urb[bulk_in_num * READS_IN_FLIGHT + j]) {
					dev_err(&interface->dev,
						"Could not allocate bulk_in_urb\n");
					goto error;
				}		
				dev->bulk_in_buffer[bulk_in_num * READS_IN_FLIGHT + j] =
					kmalloc(buffer_size, GFP_KERNEL);
				if (!dev->bulk_in_buffer[bulk_in_num * READS_IN_FLIGHT + j]) {
					dev_err(&interface->dev,
						"Could not allocate bulk_in_buffer\n");
					goto error;
				}
			}
			bulk_in_num++;
		}
		/* USB DIR OUT && BULK */
		if (!(endpoint->bEndpointAddress & 0x80) && ((endpoint->bmAttributes & 3) == 0x02)) {		
			/* we found a bulk out endpoint */
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);
			dev->bulk_out_size[bulk_out_num] = buffer_size;
			dev->bulk_out_endpointAddr[bulk_out_num] = endpoint->bEndpointAddress;
			bulk_out_num++;
		}
		/* USB DIR IN && INT */
		if ((endpoint->bEndpointAddress & 0x80) && ((endpoint->bmAttributes & 3) == 0x03)) {		
			dev->interrupt_in_endpoint[int_in_num] = endpoint;
			int_in_num++;
		}
		/* USB DIR OUT && INT */
		if (!(endpoint->bEndpointAddress & 0x80) && ((endpoint->bmAttributes & 3) == 0x03)) {		
			dev->interrupt_out_endpoint[int_out_num] = endpoint;
			int_out_num++;
		}
	}
	if (!(dev->bulk_in_endpointAddr[0] && dev->bulk_out_endpointAddr[0])) {
		err("Could not find both bulk-in and bulk-out endpoints!\n");
		goto error;
	}

	if (dev->interrupt_in_endpoint[0] == NULL) {
		dbg("interrupt in endpoint not found!\n");
	}
	if (dev->interrupt_out_endpoint[0] == NULL) {
		dbg("interrupt out endpoint not found!\n");
	}

	dev->bulk_in_num = bulk_in_num;
	dev->bulk_out_num = bulk_out_num;
	dev->interrupt_in_num = int_in_num;
	dev->interrupt_out_num = int_out_num;

	dbg("bulk_in_num: %d\n", bulk_in_num);
	dbg("bulk_out_num: %d\n", bulk_out_num);
	dbg("int_in_num: %d\n", int_in_num);
	dbg("int_out_num: %d\n", int_out_num);

	if(int_in_num != 0) {
		for(i = 0; i<int_in_num; i++) {
			in_end_size = le16_to_cpu(dev->interrupt_in_endpoint[i]->wMaxPacketSize);
			
			dbg("in_end_size: %d\n", in_end_size);
			dbg("bInterval: %d\n", dev->interrupt_in_endpoint[i]->bInterval);
			
			init_waitqueue_head(&dev->read_wait[i]);

			dev->read_buffer_primary[i] = kmalloc((4 * in_end_size), GFP_KERNEL);
			if (!dev->read_buffer_primary[i]) {
				err("Couldn't allocate read_buffer_primary\n");
				retval = -ENOMEM;
				goto error;
			}

			// debug code prime the buffer 
			memset(dev->read_buffer_primary[i], 'a', in_end_size);
			memset(dev->read_buffer_primary[i] + in_end_size, 'b', in_end_size);
			memset(dev->read_buffer_primary[i] + (2 * in_end_size), 'c', in_end_size);
			memset(dev->read_buffer_primary[i] + (3 * in_end_size), 'd', in_end_size);

			dev->read_buffer_secondary[i] = kmalloc((4 * in_end_size), GFP_KERNEL);
			if (!dev->read_buffer_secondary[i]) {
				err("Couldn't allocate read_buffer_secondary\n");
				retval = -ENOMEM;
				goto error;
			}

			// debug code prime the buffer 
			memset(dev->read_buffer_secondary[i], 'e', in_end_size);
			memset(dev->read_buffer_secondary[i] + in_end_size, 'f', in_end_size);
			memset(dev->read_buffer_secondary[i] + (2 * in_end_size), 'g', in_end_size);
			memset(dev->read_buffer_secondary[i] + (3 * in_end_size), 'h', in_end_size);

			dev->interrupt_in_buffer[i] = kmalloc(in_end_size, GFP_KERNEL);
			if (!dev->interrupt_in_buffer[i]) {
				err("Couldn't allocate interrupt_in_buffer\n");
				goto error;
			}

			// debug code prime the buffer 
			memset(dev->interrupt_in_buffer[i], 'i', in_end_size);

			dev->interrupt_in_urb[i] = usb_alloc_urb(0, GFP_KERNEL);
			if (!dev->interrupt_in_urb[i]) {
				err("Couldn't allocate interrupt_in_urb\n");
				goto error;
			}
			usb_fill_int_urb(dev->interrupt_in_urb[i], dev->udev,
				usb_rcvintpipe(dev->udev,
				dev->interrupt_in_endpoint[i]->bEndpointAddress),
				dev->interrupt_in_buffer[i],
				in_end_size,
				ch37x_interrupt_in_callback,
				dev,
				dev->interrupt_in_endpoint[i]->bInterval);
		}
	}

	if(int_out_num != 0) {
		for(i = 0; i<int_out_num; i++) {
			out_end_size = le16_to_cpu(dev->interrupt_out_endpoint[i]->wMaxPacketSize);

			dbg("out_end_size: %d\n", out_end_size);

			init_waitqueue_head(&dev->write_wait[i]);

			dev->interrupt_out_buffer[i] = kmalloc(out_end_size, GFP_KERNEL);
			if (!dev->interrupt_out_buffer[i]) {
				err("Couldn't allocate interrupt_out_buffer");
				goto error;
			}

			dev->interrupt_out_urb[i] = usb_alloc_urb(0, GFP_KERNEL);
			if (!dev->interrupt_out_urb[i]) {
				err("Couldn't allocate interrupt_out_urb\n");
				goto error;
			}
		}
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* we can register the device now, as it is ready */
	retval = usb_register_dev(interface, &ch37x_class);
	if (retval) {
		/* something prevented us from registering this driver */
		err("Not able to get a minor for this device.\n");
		usb_set_intfdata(interface, NULL);
		goto error;
	}

	/* let the user know what node this device is now attached to */
	dev_info(&interface->dev,
		"USB ch37x device now attached to ch37x-%d", 
		interface->minor);
	return 0;

error:
	if (dev)
		kref_put(&dev->kref, ch37x_delete);
	return retval;
}

static void ch37x_disconnect(struct usb_interface *interface)
{
	struct usb_ch37x *dev;
	int minor = interface->minor;
	
	info("%s is excuting\n", __func__);
	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &ch37x_class);
	
	/* prevent more I/O from starting */
	mutex_lock(&dev->io_mutex);
	dev->interface = NULL;
	mutex_unlock(&dev->io_mutex);
	
	usb_kill_anchored_urbs(&dev->submitted);

	/* decrement our usage count */
	kref_put(&dev->kref, ch37x_delete);

	dev_info(&interface->dev, "USB ch37x #%d now disconnected", minor);
}

static void ch37x_draw_down(struct usb_ch37x *dev)
{
	int time;
	int i;

	info("%s - excuting!\n", __func__);
	time = usb_wait_anchor_empty_timeout(&dev->submitted, 1000);
	if (!time)
		usb_kill_anchored_urbs(&dev->submitted);
	for( i = 0; i < READS_IN_FLIGHT; i++ ) {
		usb_kill_urb(dev->bulk_in_urb[i]);
		dev->ongoing_read[i] = 0;
	}
}

static int ch37x_suspend(struct usb_interface *intf, pm_message_t message)
{
	struct usb_ch37x *dev = usb_get_intfdata(intf);

	info("%s - excuting!\n", __func__);
	if (!dev)
		return 0;
	ch37x_draw_down(dev);
	return 0;
}

static int ch37x_resume(struct usb_interface *intf)
{
	info("%s - excuting!\n", __func__);
	return 0;
}

static int ch37x_pre_reset(struct usb_interface *intf)
{
	struct usb_ch37x *dev = usb_get_intfdata(intf);
	info("%s - excuting!\n", __func__);

	mutex_lock(&dev->io_mutex);
	ch37x_draw_down(dev);

	return 0;
}

static int ch37x_post_reset(struct usb_interface *intf)
{
	struct usb_ch37x *dev = usb_get_intfdata(intf);
	info("%s - excuting!\n", __func__);

	/* we are sure no URBs are active - no locking needed */
	dev->errors = -EPIPE;
	mutex_unlock(&dev->io_mutex);

	return 0;
}

static struct usb_driver ch37x_driver = {
	.name =		"ch37x",
	.probe =	ch37x_probe,
	.disconnect =	ch37x_disconnect,
	.suspend =	ch37x_suspend,
	.resume =	ch37x_resume,
	.pre_reset =	ch37x_pre_reset,
	.post_reset =	ch37x_post_reset,
	.id_table =	ch37x_table,
	.supports_autosuspend = 1,
};

static int __init usb_ch37x_init(void)
{
	int ret;
	ret = usb_register(&ch37x_driver);
	if(ret)
		err("usb_register failed. Error number: %d\n", ret);
	return ret;
}

static void __exit usb_ch37x_exit(void)
{
	usb_deregister(&ch37x_driver);
}

module_init(usb_ch37x_init);

module_exit(usb_ch37x_exit);

MODULE_LICENSE("GPL");

